from types import LambdaType
from typing import Tuple
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as colors
from pybasicDTW.tools import Tools

class DTW(Tools):
    def __init__(self, metric:LambdaType = lambda x,y: np.square(x-y) ,stepPattern:np.ndarray = np.array([(1,1),(1,0),(0,1)]), stepWeights:np.ndarray = np.array([])) -> None:
        """Initalise dtw core with parameters"""
        # check stepPattern dimension
        if stepPattern.shape[0] == 0: raise ValueError("No stepPattern given.")
        if stepPattern.shape[1] != 2: raise ValueError("Each stepPattern must only contain two parameters, e.g (1,1).")
        # default stepWeights if not given
        if stepWeights.size == 0: stepWeights = np.ones((stepPattern.shape[0],))
        # check stepWeights if given
        if stepWeights.size > 0:
            if stepWeights.ndim != 1: raise ValueError("stepWeights must be one dimensional.")
            if stepWeights.shape[0] != stepPattern.shape[0]: raise ValueError(f"stepWeight dimension unequal: stepWeight:{stepWeights.shape[0]} != stepPattern:{stepPattern.shape[0]}")
        else: raise TypeError("stepWeights invalid type.")
        # check metric
        if metric.__code__.co_argcount != 2: raise TypeError("Metric lambda must have two inputs.")
        # Save all parameters to private variable
        self._stepPattern = stepPattern
        self._stepWeights = stepWeights
        self._metric = metric

    def CostMatrix(self, x:np.ndarray, y:np.ndarray, dimWeights:np.ndarray = np.array([]), subsequence:bool=False):
        """Calculate the local and accumulated cost of two sequences of varying dimensions."""
        # convert one dimensional sequency into correct format, each element represents some time t.
        if x.ndim == 1: x = np.array([[xElement] for xElement in x])
        if y.ndim == 1: y = np.array([[yElement] for yElement in y])
        # validate x and y parameters
        if x.shape[1] != y.shape[1]: raise ValueError("x and y must have equal dimensions along axis=1 ")
        # default dimWeights
        if dimWeights.size == 0: dimWeights = np.ones((x.shape[1]))
        # check dimWeights dimensions are equal
        if dimWeights.size > 0:
            if dimWeights.ndim != 1: raise ValueError("dimensionWeights must be one dimensional.")
            if dimWeights.shape[0] != x.shape[1]: raise ValueError("dimWeights and x,y must have equal dimensions along axis=1")

        # local cost matrix: vectorisation/broadcasting
        if x.shape[0] > y.shape[0]: y = y[:, np.newaxis]
        if y.shape[0] > x.shape[0] or y.shape[0] == x.shape[0]: x = x[:, np.newaxis]
        metricMatrix = self._metric(x,y) * dimWeights
        localCost = np.sum(metricMatrix, axis=2)
        # accumulated cost matrix: for loop
        # 1. create empty extended accumulated cost matrix dependent on stepPattern
        maxStepPattern = (self._stepPattern[np.argmax(self._stepPattern[:,0])][0],  self._stepPattern[np.argmax(self._stepPattern[:,1])][1])
        accumulatedCost = np.zeros(shape=(localCost.shape[0]+maxStepPattern[0],localCost.shape[1]+maxStepPattern[1]))
        zeroIndex = maxStepPattern
        # 2. Set default values in extended quadrants of the matrix
        accumulatedCost[:,0:zeroIndex[1]] = np.inf
        accumulatedCost[0:zeroIndex[0],:] = np.inf
        accumulatedCost[0:zeroIndex[0],0:zeroIndex[1]] = 0
        # 3. Ammend default values if subsequence to give start point flexibility
        if subsequence: accumulatedCost[:zeroIndex[0],zeroIndex[0]:] = 0
        # 4. calculate accumulated costs
        for r in range(zeroIndex[0], accumulatedCost.shape[0]):
            for c in range(zeroIndex[1], accumulatedCost.shape[1]):
                stepChoices = np.array([accumulatedCost[r-step[0], c-step[1]] for step in self._stepPattern])
                accumulatedCost[r,c] = stepChoices[self.LexiArgMinIndex(stepChoices)] + localCost[r-zeroIndex[0], c-zeroIndex[1]]
        
        return accumulatedCost[zeroIndex[0]:, zeroIndex[1]:], localCost

    def OptimalWarpingPath(self, accCost:np.ndarray, localCost:np.ndarray, endIndex:Tuple=None):
        """Calculate the optimal warping path, working backwards from the last point to the first point."""
        # default end point for classical dtw
        forceStartPoint = False
        if not isinstance(endIndex, Tuple): 
            endIndex = (accCost.shape[0]-1, accCost.shape[1]-1)
            forceStartPoint  = True

        path = [endIndex]
        # iterate to find optimal path
        startReached = False
        while not startReached:
            # step choices
            stepChoices = []
            for step, weight in zip(self._stepPattern, self._stepWeights):
                row = path[-1][0] - step[0] if (path[-1][0] - step[0]) >= 0 else 0
                col = path[-1][1] - step[1] if (path[-1][1] - step[1]) >= 0 else 0
                stepChoices.append( (accCost[row,col]*weight, (row,col)) )
            stepChoices = np.array(stepChoices, dtype="object")
            # ensure no steps are repeated
            stepChoices[stepChoices == np.array([(accCost[path[-1]] ,path[-1])], dtype="object")] = np.inf
            # optimal step
            optimalStep = stepChoices[self.LexiArgMinIndex(stepChoices[:,0])][1]
            path.append(optimalStep)
            # loop terminations
            if forceStartPoint and path[-1] == (0,0): startReached = True
            if not forceStartPoint and path[-1][0] == 0: startReached = True
        # calculate total dtw cost
        totalCost = np.sum(np.array([localCost[point] for point in path]))
        return np.array(path), totalCost



if __name__ == "__main__":
    dtw = DTW(stepPattern=np.array([(2,2),(2,0),(0,2)]))

    x = np.array([1,2,3,4,5])
    y = np.array([10,10,10])

    accumulatedCost, localCost = dtw.CostMatrix(x,y, subsequence=True)
    path, totalCost = dtw.OptimalWarpingPath(accumulatedCost, localCost, endIndex=(accumulatedCost.shape[0]-1, accumulatedCost.shape[1]-1))

    dtw.PlotCost(accumulatedCost, path, title=f"Accum Cost + path, cost={totalCost}")