from typing import Tuple
from pybasicDTW.dtw import DTW
from types import LambdaType
import numpy as np
import logging

class Subsequence(DTW):
    def __init__(self, x:np.ndarray, y:np.ndarray, dimWeights = np.array([]), metric:LambdaType = lambda x,y: np.square(x-y) ,stepPattern:np.ndarray = np.array([(1,1),(1,0),(0,1)]), stepWeights:np.ndarray = np.array([])) -> None:
        # Initalise DTW
        super().__init__(metric= metric, stepPattern=stepPattern, stepWeights=stepWeights)
        # Calculate cost matrix
        self.__ACost, self.__LCost = self.CostMatrix(x=x, y=y, dimWeights=dimWeights, subsequence=True)
        # copy top row of accumulated cost matrix
        self.__EndPoints = np.copy(self.__ACost[self.__ACost.shape[0]-1,:])
    
    @property
    def AccumulatedCost(self):
        return self.__ACost

    @property
    def LocalCost(self):
        return self.__LCost

    def FindMatch(self, exclusionType:str = "LocalDistance", LocalMaxDist:int=1, invert=True) -> Tuple:
        """Find SubSequence Match

        Args:
            exclusionType (str, optional): [description]. Defaults to "LocalDistance".
            LocalMaxDist ([type], optional): [description]. Defaults to None.

        Raises:
            TypeError: [description]

        Returns:
            Tuple: [description]
        """
        # check correct excludion type present
        if exclusionType != "LocalDistance" and exclusionType != "LocalMaximum" : raise TypeError("Please select either 'LocalDistance' or 'LocalMaximum exclusion types.")
        # if no more matches can be found
        if np.all(self.__EndPoints == np.inf):
            logging.warn("No more matches can be found!")
            return None, -1
        # find minimum end point
        minIndex = self.LexiArgMinIndex(self.__EndPoints, invert=invert)
        # find optimal path and total cost
        path, totalCost = self.OptimalWarpingPath(self.__ACost, self.__LCost, (self.__LCost.shape[0]-1, minIndex))
        # neighbouring point exclusion type
        if exclusionType == "LocalDistance":
            # soft error for distance larger than error
            if LocalMaxDist > self.__EndPoints.shape[0]: logging.warn("LocalDistanceExclusion:   Distance exceeds array length, no end points to select after this!")
            self.LocalDistanceExclusion(minIndex, self.__EndPoints, LocalMaxDist)
        if exclusionType == "LocalMaximum": self.LocalMaximumExclusion(minIndex, self.__EndPoints)
        return path, totalCost

    @classmethod
    def LocalDistanceExclusion(cls, targetIndex:int, searchArray:np.ndarray, distance:int) -> None:
        """Exclude points within set distance from center.

        Args:
            targetIndex (int): [description]
            searchArray (np.ndarray): FLOAT Values.
            distance (int): [description]
        """
        arrayLength = searchArray.shape[0]

        # exclude points to the left
        leftStart = targetIndex - distance
        # exclude points to the right
        if leftStart < 0: leftStart = 0
        leftEnd = targetIndex
        searchArray[leftStart:leftEnd+1] = np.inf

        # exclude points to the right
        rightStart = targetIndex + 1
        rightEnd = targetIndex + distance
        if rightEnd > arrayLength: rightEnd = arrayLength - 1
        searchArray[rightStart:rightEnd+1] = np.inf

    @classmethod
    def LocalMaximumExclusion(xls, targetIndex:int, searchArray:np.ndarray) -> None:
        """Find local maximum around "center" point and set to neighbours to infinity.

        Args:
            targetIndex (int): index of minimum point to look around.
            searchArray (np.ndarray): reference to list containing all points.
        """

        # Find maximum to the left if targetIndex is not start
        if targetIndex != 0:
            for i in range(targetIndex-1,-1,-1):
                searchArray[i] = np.inf
                if i != 0 and searchArray[i] > searchArray[i-1]: break
        # Find maximum to the right if targetIndex is not end
        if targetIndex != (searchArray.shape[0] - 1):
            for i in range(targetIndex+1,searchArray.shape[0],1):
                searchArray[i] = np.inf
                if i != searchArray.shape[0]-1 and searchArray[i] > searchArray[i+1]: break

        searchArray[targetIndex] = np.inf
