import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as colors

class Tools:
    def LexiArgMinIndex(self, list: np.ndarray, invert:bool = False) -> int:
        """ Get the lexiographical minimum in a 1D array.

        Lexiographical: generalisation of alphabetical order of the dictionaries to sequence of ordered elements.        

        If invert is False: if there are two minimmums, the minimum with the smallest index will be chosen.
        If invert is True : if there are two minimums, the minimum with the largest index will be chosen.

        Args:
            list (np.ndarray): 1D numpy array of elements.
            invert (bool, optional): Invert lexiographical order of non-unique minimums. Defaults to False.

        Returns:
            int: Index of lexiographical minimum value in list.
        """
        # find all occurrences which has the lowest cost and flatten to 1D array
        minimums = np.argwhere(list == np.amin(list)).flatten()
        # if only one minimum found (UNIQUE)
        if len(minimums) == 1: return minimums[0]
        # if more than one minimum found (NON-UNIQUE)
        if invert: return np.amax(minimums)
        return np.amin(minimums)

    def PlotCost(self, matrix:np.ndarray, warpingPath:np.ndarray=None, title:str = "", xLabel:str="", yLabel:str="", colorLabel:str="") -> None:
        Fig, Ax = plt.subplots()
        # custom color to highlight lower cost areas more
        cmap = colors.LinearSegmentedColormap.from_list('customHeat',['#00ff00','#00ccff',"#0000cc"],gamma=.3)
        colorMap = Ax.matshow(matrix,origin="lower", aspect="auto", cmap=cmap)
        Ax.set_axis_off()
        Fig.colorbar(colorMap, ax=Ax, label=colorLabel)
        # ylabel
        if isinstance(yLabel, str): Ax.set_ylabel(yLabel)
        if isinstance(xLabel, str): Ax.set_xlabel(xLabel)
        if isinstance(title, str): 
            Ax.set_title(title, y=1.0,)
        # for path overaly
        if isinstance(warpingPath,np.ndarray):
            Ax.plot([x[1] for x in warpingPath], [y[0] for y in warpingPath],color="red", label="Warpped Optimal Path")
        plt.show()
